import logging

from ..appproperty import HEA_DB
from .. import response
from ..heaobjectsupport import new_heaobject
from ..aiohttp import StreamReaderWrapper
from heaobject.error import DeserializeException
from aiohttp.web import Request, Response
from typing import Type, IO, Optional
from heaobject.root import HEAObject


async def get(request: Request, collection: str) -> Response:
    """
    Gets the HEA object with the specified id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :return: a Response with the requested HEA object or Not Found.
    """
    result = await request.app[HEA_DB].get(request, collection, var_parts='id')
    return await response.get(request, result)


async def get_content(request: Request, collection: str) -> Response:
    """
    Gets the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :return: a Response with the requested HEA object or Not Found.
    """
    out = await request.app[HEA_DB].get_content(request, collection, var_parts='id')
    if out is not None:
        return await response.get_streaming(request, StreamReaderWrapper(out), 'text/plain')
    else:
        return response.status_not_found()


async def get_by_name(request: Request, collection: str) -> Response:
    """
    Gets the HEA object with the specified name.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :return: a Response with the requested HEA object or Not Found.
    """
    result = await request.app[HEA_DB].get(request, collection, var_parts='name')
    return await response.get(request, result)


async def get_all(request: Request, collection: str) -> Response:
    """
    Gets all HEA objects.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :return: a Response with a list of HEA object dicts.
    """
    result = await request.app[HEA_DB].get_all(request, collection)
    return await response.get_all(request, result)


async def opener(request: Request, collection: str) -> Response:
    """
    Gets choices for opening an HEA desktop object's content.

    :param request: the HTTP request. Required. If an Accepts header is provided, MIME types that do not support links
    will be ignored.
    :param collection: the Mongo collection name. Required.
    :return: a Response object with status code 300, and a body containing the HEA desktop object and links
    representing possible choices for opening the HEA desktop object; or Not Found.
    """
    result = await request.app[HEA_DB].get(request, collection, var_parts='id')
    return await response.get_multiple_choices(request, result)


async def post(request: Request, collection: str, type_: Type[HEAObject], default_content: Optional[IO] = None) -> Response:
    """
    Posts the provided HEA object.

    :param request: the HTTP request.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param default_content: an optional blank document or other default content as a file-like object. This must be not-None
    for any microservices that manage content.
    :return: a Response object with a status of Created and the object's URI in the
    """
    try:
        obj = await new_heaobject(request, type_)
        result = await request.app[HEA_DB].post(request, obj, collection, default_content)
        return await response.post(request, result, collection)
    except DeserializeException:
        return response.status_bad_request()


async def put(request: Request, collection: str, type_: Type[HEAObject]) -> Response:
    """
    Updates the HEA object with the specified id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :return: a Response object with a status of No Content or Not Found.
    """
    try:
        obj = await new_heaobject(request, type_)
        if request.match_info['id'] != obj.id:
            return response.status_bad_request()
        result = await request.app[HEA_DB].put(request, obj, collection)
        return await response.put(result.matched_count if result else False)
    except DeserializeException:
        return response.status_bad_request()


async def put_content(request: Request, collection: str, type_: Type[HEAObject]) -> Response:
    """
    Updates the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :return: a Response object with a status of No Content or Not Found.
    """
    try:
        result = await request.app[HEA_DB].put_content(request, collection)
        return await response.put(result)
    except DeserializeException:
        return response.status_bad_request()


async def delete(request: Request, collection: str) -> Response:
    """
    Deletes the HEA object with the specified id and any associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :return: No Content or Not Found.
    """
    result = await request.app[HEA_DB].delete(request, collection)
    return await response.delete(result.deleted_count if result else False)



"""
Form data representor. It only supports the parse method (form data to NVPJSON), and it does not support data values
with nested JSON objects.
"""

from urllib.parse import parse_qsl
from .error import ParseException
from .representor import Representor, Link
from aiohttp.web import Request
from typing import Mapping, Any, Union, Dict, List, Callable
import json
from yarl import URL
import logging


MIME_TYPE = 'application/x-www-form-urlencoded'


class XWWWFormURLEncoded(Representor):
    MIME_TYPE = MIME_TYPE

    async def formats(self, request: Request,
                      wstl_obj: Union[List[Dict[str, Any]], Dict[str, Any]],
                      dumps=json.dumps,
                      link_callback: Callable[[int, Link], None] = None) -> bytes:
        raise NotImplementedError

    async def parse(self, request: Request) -> Mapping[str, Any]:
        """
        Parses an HTTP POST request containing form data into a name-value pair dict-like object.

        :param request: the HTTP request. Cannot be None.
        :return: the data section of the Collection+JSON document transformed into a dict-like object.
        """
        _logger = logging.getLogger(__name__)
        try:
            txt = await request.text()
            _logger.debug('Parsing %s', txt)
            result = parse_qsl(txt, strict_parsing=True)
            _logger.debug('Parsed to %s', result)
            d = {k: v for k, v in result}
            _logger.debug('Returning %s', d)
            return d
        except ValueError as e:
            raise ParseException from e





from testcontainers.general import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready
import requests


@wait_container_is_ready()
def get_exposed_port(container: DockerContainer, port: int):
    return container.get_exposed_port(port)


@wait_container_is_ready()
def wait_for_status_code(url, status):
    return requests.get(url).status_code == status


def get_bridge_ip(container: DockerContainer) -> str:
    """
    Returns the IP address of the container on the default bridge network.
    :param container: a docker container.
    :return: an IP address.
    """
    return container.get_docker_client().bridge_ip(container.get_wrapped_container().id)



"""
Functions for working with pipes in asyncio.
"""

from asyncio import StreamReader, StreamWriter, StreamReaderProtocol, get_running_loop
import os
from typing import IO, Union


def pipe():
    read_fileno, write_fileno = os.pipe()
    read_fd = os.fdopen(read_fileno, 'rb')
    write_fd = os.fdopen(write_fileno, 'wb')
    return read_fd, write_fd


async def pipe_stream():
    read_fd, write_fd = pipe()
    reader = await _create_pipe_stream_reader(read_fd)
    writer = await _create_pipe_stream_writer(write_fd)
    return reader, writer


async def transfer(input_: Union[IO, StreamReader], output: IO) -> None:
    """
    Transfer bytes from an input file-like object or asyncio.StreamReader to an output file-like object.
    :param input_: the input file-like object or stream reader.
    :param output: the output file-like object.
    """
    if os.name == 'nt':
        await _piped_nt(input_, output)
    else:
        await _piped_other(input_, output)


async def _piped_nt(input_: Union[IO, StreamReader], output: IO) -> None:
    read_fd, write_fd = pipe()
    get_running_loop().run_in_executor(None, _do_write, input_, write_fd, 1024)
    _do_reading(read_fd, output)


async def _piped_other(input_: Union[IO, StreamReader], output: IO) -> None:
    reader, writer = await pipe_stream()
    get_running_loop().create_task(_do_write_async(input_, writer, 1024))
    await _do_reading_async(reader, output)


def _do_reading(reader: IO, output: IO) -> None:
    while chunk := reader.read(2 ** 16):
        output.write(chunk)


async def _do_reading_async(reader: StreamReader, output: IO) -> None:
    while not reader.at_eof():
        chunk = await reader.read(2 ** 16)
        output.write(chunk)


async def _create_pipe_stream_reader(read_fd: IO) -> StreamReader:
    loop = get_running_loop()
    reader = StreamReader()
    read_protocol = StreamReaderProtocol(reader)
    read_transport, _ = await loop.connect_read_pipe(lambda: read_protocol, read_fd)
    return reader


async def _create_pipe_stream_writer(write_fd: IO) -> StreamWriter:
    loop = get_running_loop()
    write_protocol = StreamReaderProtocol(StreamReader())
    write_transport, _ = await loop.connect_write_pipe(lambda: write_protocol, write_fd)
    return StreamWriter(write_transport, write_protocol, None, loop)


async def _do_write_async(fd: Union[IO, StreamReader], writer: StreamWriter, n: int = None):
    """
    Writes the contents of a file-like object fd to a stream writer, and closes the writer.

    :param fd: the file-like object supporting read().
    :param writer: the stream writer.
    :param n: read up to size bytes from the object.
    """
    try:
        if isinstance(fd, StreamReader):
            while not fd.at_eof():
                writer.write(fd.read(n))
                await writer.drain()
        else:
            while chunk := fd.read(n):
                writer.write(chunk)
                await writer.drain()
        writer.close()
        await writer.wait_closed()
        writer = None
    finally:
        if writer is not None:
            try:
                writer.close()
                await writer.wait_closed()
            except OSError:
                pass


def _do_write(fd: Union[IO, StreamReader], writer: IO, n: int = None):
    """
    Writes the contents of a file-like object fd to another file-like object, and closes the latter.

    :param fd: the file-like object supporting read().
    :param writer: the stream writer.
    :param n: read up to size bytes from the object.
    """
    try:
        while chunk := fd.read(n):
            writer.write(chunk)
        writer.close()
    finally:
        if not writer.closed:
            try:
                writer.close()
            except OSError:
                pass
